local tonumber = tonumber
local select = select
local ipairs = ipairs
local pairs = pairs
local ceil = ceil
local floor = floor
local abs = abs
local getglobal = getglobal
local UnitPowerType = UnitPowerType
local UnitHealth = UnitHealth
local UnitMana = UnitMana
local GetActionInfo = GetActionInfo
local GetActionCount = GetActionCount
local IsConsumableAction = IsConsumableAction
local GetItemCount = GetItemCount
local UnitIsDeadOrGhost = UnitIsDeadOrGhost

ltc = AceLibrary("AceAddon-2.0"):new("AceEvent-2.0", "AceHook-2.1", "AceConsole-2.0", "AceDB-2.0")
local Gratuity = AceLibrary("Gratuity-2.0")

local Debug
local Plugins = {}
local CostList = setmetatable({}, { __mode = "kv" })
local ReagentList = {}
local CountList = {}

local L = AceLibrary("AceLocale-2.2"):new("lern2count")

local CostTable = {
	{ type = 0, base = MANA_COST, timed = MANA_COST_PER_TIME }, 
	{ type = 1, base = RAGE_COST, timed = RAGE_COST_PER_TIME },
	{ type = 2, base = FOCUS_COST, timed = FOCUS_COST_PER_TIME },
	{ type = 3, base = ENERGY_COST, timed = ENERGY_COST_PER_TIME },
	{ type = 4, base = HEALTH_COST, timed = HEALTH_COST_PER_TIME },
}

local ltc = ltc

ltc:RegisterDB("lern2countDB")
ltc:RegisterDefaults('profile', {
	colors = {
		[0] = "88aaff",
		[1] = "ffff33",
		[2] = "ff8033",
		[3] = "ffff33",
		[4] = "ff0033",
		R   = "00ff33",
		D   = "ffffff",
	},
	sizes = {
		[0] = 14,
		[1] = 14,
		[2] = 14,
		[3] = 14,
		[4] = 14,
		R   = 14,
		D   = 14,
	},
	hidden = {
		[0] = false,
		[1] = false,
		[2] = false,
		[3] = false,
		[4] = false,
		R   = false,
		D   = false,
	},
})

local function getRGB(hex)
	local r = tonumber(hex:sub(1, 2), 16) / 255
	local g = tonumber(hex:sub(3, 4), 16) / 255
	local b = tonumber(hex:sub(5, 6), 16) / 255
	return r, g, b
end

local function setRGB(r, g, b)
	local hex = ("%02x%02x%02x"):format(floor(r * 255), floor(g * 255), floor(b * 255))
	return hex
end

function ltc:OnInitialize()
	local nameMap = {
		[0] = L["mana"],
		[1] = L["rage"],
		[2] = L["focus"],
		[3] = L["energy"],
		[4] = L["health"],
		R   = L["reagent"],
		D   = L["count"],
	}
	local options = {
		desc = L["An addon to show reagent count on action buttons"],
		type = "group",
		args = {
			colors = {
				name = L["colors"],
				desc = L["Color of the text for the different elements"],
				type = "group",
				args = {},
			},
			sizes = {
				name = L["sizes"],
				desc = L["Size of the text for the different elements"],
				type = "group",
				args = {},
			},
			hidden = {
				name = L["hide"],
				desc = L["Hide the text for the different elements"],
				type = "group",
				args = {},
			},
		}
	}
	for k, v in pairs(nameMap) do
		local index = k
		options.args.colors.args[v] = {
			name = v,
			desc = L["Color for %s"]:format(v),
			type = "color",
			get = function () return getRGB(self.db.profile.colors[index]) end,
			set = function (r, g, b) self.db.profile.colors[index] = setRGB(r, g, b) ; self:UpdateAllCounts() end,
		}
		options.args.sizes.args[v] = {
			name = v,
			desc = L["Fontsize for %s"]:format(v),
			type = "range",
			min = 6,
			max = 20,
			step = 1,
			get = function () return self.db.profile.sizes[index] end,
			set = function (v) self.db.profile.sizes[index] = v ; self:UpdateAllCounts() end,
		}
		options.args.hidden.args[v] = {
			name = v,
			desc = L["Hide %s"]:format(v),
			type = "toggle",
			get = function () return self.db.profile.hidden[index] end,
			set = function (v) self.db.profile.hidden[index] = v ; self:CheckTooltipOption(); self:UpdateAllCounts() end,
		}
	end
	
	self:RegisterChatCommand(L["CONSOLE_CMDS"], options)
end

do 
	local checktypes = { 0, 1, 2, 3, 4, "R" }
	function ltc:CheckTooltipOption()
		if GetCVar("UberTooltips") ~= "0" then return end
		for _,v in ipairs(checktypes) do
			if not self.db.profile.hidden[v] then
				self:Print(L["UberTooltips are required for lern2count correctly displaying information, activating it"])
				SetCVar("UberTooltips", "1")
				break
			end
		end
	end
end

local DelayEvent = "ltcDelayEvent"
function ltc:OnEnable()
	self:RegisterBucketEvent("UNIT_HEALTH", 0.2, "PowerUpdate")
	self:RegisterBucketEvent("UNIT_MANA", 0.2, "PowerUpdate")
	self:RegisterBucketEvent("UNIT_RAGE", 0.2, "PowerUpdate")
	self:RegisterBucketEvent("UNIT_ENERGY", 0.2, "PowerUpdate")
	self:RegisterBucketEvent("UNIT_AURA", 0.2)
	self:RegisterBucketEvent("BAG_UPDATE", 1.0)
	self:RegisterEvent("ACTIONBAR_SLOT_CHANGED","BarUpdate")
	self:RegisterEvent("ACTIONBAR_HIDEGRID","BarUpdate")
	self:RegisterEvent("ACTIONBAR_PAGE_CHANGED","BarUpdate")
	self:RegisterEvent("ACTIONBAR_UPDATE_USABLE","BarUpdate")
	self:RegisterEvent("ACTIONBAR_PAGE_CHANGED","BarUpdate")
	self:RegisterEvent("UPDATE_BONUS_ACTIONBAR","BarUpdate")
	self:RegisterEvent("UPDATE_SHAPESHIFT_FORMS","BarUpdate")
	self:RegisterEvent("PET_BAR_HIDEGRID","BarUpdate")
	self:RegisterEvent("PET_BAR_UPDATE","BarUpdate")
	self:RegisterEvent("PET_BAR_UPDATE_COOLDOWN","BarUpdate")

	self:ScheduleEvent(DelayEvent, self.DelayedUpdate, 1, self) --Delay initializing the addon a second, thats enough for everything to finish what its doing (I hope)

	self:RegisterEvent("PLAYER_REGEN_DISABLED")
	self:RegisterEvent("PLAYER_REGEN_ENABLED")
	
	self:SecureHook("ActionButton_UpdateCount")

end

function ltc:PLAYER_REGEN_DISABLED()
	self.combat = true
	self:UpdateAllCounts()
end

function ltc:PLAYER_REGEN_ENABLED()
	self.combat = false
	self:UpdateAllCounts()
end

local GratuityHandler = {
	action = function (id) Gratuity:SetAction (id) end,
	pet = function (id) Gratuity:SetPetAction (id) end,
	shapeshift = function (id) Gratuity:SetShapeshift (id) end,
}

function ltc:UpdateCount(frame, info)	
	local ptype = UnitPowerType("player")

	GratuityHandler[info.type](info.id)
	
	local ctype, cost
	do
		local cl = CostList[frame]
		if cl then
			ctype = cl.type
			cost = cl.cost
		end
	end
	
	local amount
	if info.type == "pet" then
		amount = UnitMana("pet")
	elseif ctype == 4 then
		amount = UnitHealth("player")
	else
		amount = UnitMana("player")
	end
	
	local rcount
	do
		local reagent = ReagentList[frame]
		if reagent then
			rcount = GetItemCount(reagent)
		end
	end
	
	local pcount
	if cost then
		if ctype == 3 then -- Energy
			if ctype == ptype then
				pcount = ceil( (cost - amount) / 20 )
				if pcount <= 0 or UnitIsDeadOrGhost("player") then 
					pcount = nil 
				end
			else
				pcount = nil
			end
		else
			pcount = floor(amount / cost)
		end
		if ctype == 1 and (pcount == 0 or ptype ~= 1) then
			pcount = nil
		end
	end

	local value
	if self.combat then
		if rcount and (not pcount or rcount < pcount) then
			value =  rcount
			ctype = "R"
		else
			value = pcount
		end
	else
		if rcount then
			value = rcount
			ctype = "R"
		else
			value = pcount
		end
	end
		
	if not value and info.type == "action" and IsConsumableAction (info.id) then
		value = GetActionCount (info.id)
		ctype = "D"
	end
	
	local p = self.db.profile
	if value and ctype and not p.hidden[ctype] then
		frame:SetText("|cff"..p.colors[ctype]..value.."|r")
		local new_height = p.sizes[ctype]
		
		local font, height, flag = frame:GetFont()
		if abs(height - new_height) > .5 then
			frame:SetFont(font, new_height, flag)
		end
	else
		frame:SetText()
	end

	if Debug then
		Debug(frame:GetName(), info.type, info.id, ptype, ctype, value)
	end
end

function ltc:GetCountList()
	return CountList
end

function ltc:UpdateAllCounts()
	for frame, info in pairs(CountList) do
		self:UpdateCount (frame, info)
	end
end

function ltc:ActionButton_UpdateCount()
	local frame = getglobal(this:GetName().."Count")
	local info = CountList[frame]
	if info then
		self:UpdateCount (frame, info)
	end
end

function ltc:DelayedUpdate()

	self:CheckTooltipOption()
	self:CountListUpdate()
	self:CostAndReagentListUpdate()
	self:UpdateAllCounts()
end

function ltc:RegisterPlugin(func)
	Plugins[#Plugins + 1] = func
end

local function AddFrameCount(frame, type, id)
	if type ~= "action" or GetActionInfo(id) then
		local t = CountList[frame]
		if not t then
			t = {}
			CountList[frame] = t
		end
		t.type = type
		t.id = id
		t.hidden = nil
	end
end

function ltc:CountListUpdate()
	for _, v in pairs(CountList) do
		v.hidden = true
	end

	for _, func in ipairs(Plugins) do
	 	func (AddFrameCount)
	end
end

function ltc:CostAndReagentListUpdate()
	for frame, info in pairs(CountList) do
		if not info.hidden then
			GratuityHandler[info.type](info.id)
			if Gratuity:NumLines() >= 2 then
			
				for _,c in pairs(CostTable) do
					local cost = 
						Gratuity:FindDeformat(c.base, 2, 2, false, true) or 
						Gratuity:FindDeformat(c.timed, 2, 2, false, true)

					if cost then
						cost = tonumber(cost)
						local t = CostList[frame]
						if not t then
							t = {}
							CostList[frame] = t
						end
						t.type = c.type
						t.cost = cost
						break
					end
				end
				
				local count
				local reagent = select(3, Gratuity:Find(SPELL_REAGENTS.."(.+)", 2, 99, false, true))
				if reagent then
					if reagent:sub(0, 2) == "|c" then
						reagent = reagent:sub(11, -3)
					end
				end
				ReagentList[frame] = reagent
			end
		end
	end
end

function ltc:BAG_UPDATE()
	self:UpdateAllCounts()
end

function ltc:PowerUpdate(units)
	if units.player or units.pet then
		self:UpdateAllCounts()
	end
end

function ltc:BarUpdate()
	self:ScheduleEvent(DelayEvent, self.DelayedUpdate, 0.2, self)
end

function ltc:UNIT_AURA(units)
	if units.player then
		self:BarUpdate()
	end
end

if false then
	local data = {}
	local Tablet = AceLibrary("Tablet-2.0")
	local entries = {}
	local typeNames = {
		[0] = "Mana",
		[1] = "Rage",
		[2] = "Focus",
		[3] = "Energy",
		[4] = "Health",
		R = "Reagent",
		D = "Count",
	}
	Debug = function (name, type, id, ptype, ctype, count)
		entries[name] = { type, id, ptype, ctype, count }
		ltc:TriggerEvent("lern2count_DebugUpdate")
	end
	
	function ltc:lern2count_DebugUpdate()
		Tablet:Refresh("lern2count")
	end
	
	ltc:RegisterEvent("lern2count_DebugUpdate", 1)
	
	local Children = function ()
		Tablet:SetTitle("lern2count Debug")
		local cat = Tablet:AddCategory(
			"id", 1,
			"columns", 6,
			"text", "name",
			"text2", "type",
			"text3", "id",
			"text4", "ptype",
			"text5", "ctype",
			"text6", "count"
		)
		local names = {}
		for n in pairs(entries) do
			table.insert(names, n)
		end
		table.sort(names)
		
		for _,n in ipairs(names) do
			local v = entries[n]
			cat:AddLine(
				"text", n,
				"text2", v[1],
				"text3", v[2],
				"text4", v[3],
				"text5", v[4] and typeNames[v[4]] or "nil",
				"text6", v[5]
			)
		end
	end
	
	Tablet:Register("lern2count",
		"data", data,
		"children", Children,
		"showTitleWhenDetached", true,
		"dontHook", true,
		"cantAttach", true,
		"showHintWhenDetached", true
	)
	Tablet:Detach("lern2count")
end

